from setuptools import setup

setup(name='zmqhelper',
      version='0.1',
      description='Convenient server and client classes for pyZMQ',
      url='',
      author='Krister Shalm, Joseph Cavanagh',
      author_email='lks@nist.gov',
      license='MIT',
      packages=['zmqhelper'],
      include_package_data=True,
      zip_safe=False)
